@@ -115,21 +115,56 @@ class BrowserIntegrationFlatpakIntegrator : public KDEDModule
     Q_CLASSINFO("D-Bus Interface", "org.kde.plasma.browser.integration.FlatpakIntegrator")
 
 public:
+    /** Enum for different browser types, i.e. which browser they are based on. */
+    enum BrowserBase {
+        Firefox,
+        Chrome,
+        Chromium,
+    };
+    Q_ENUM(BrowserBase)
+
+    /** Browser information structure to hold browser-specific configuration. */
+    struct BrowserInfo {
+        BrowserBase base;
+
+        /** The browser's Flatpak id, e.g. "org.mozilla.firefox". */
+        QString id;
+
+        /** The directory the browser expects its native messaging hosts to be in. */
+        QString nativeMessagingHostsDir;
+    };
+
     BrowserIntegrationFlatpakIntegrator(QObject *parent, const QList<QVariant> &)
         : KDEDModule(parent)
     {
-        auto flatpak = new QProcess(this);
-        connect(flatpak, qOverload<int, QProcess::ExitStatus>(&QProcess::finished), this, [flatpak](int, QProcess::ExitStatus) {
-            flatpak->deleteLater();
-        });
-        flatpak->start(u"flatpak"_s, {u"override"_s, u"--user"_s, u"--talk-name=org.kde.plasma.browser.integration"_s, u"org.mozilla.firefox"_s});
+        // List of Flatpak browsers we support.
+        const QList<BrowserInfo> supportedBrowsers = {
+            {BrowserBase::Firefox, u"org.mozilla.firefox"_s, u"/.mozilla/native-messaging-hosts"_s},
+            {BrowserBase::Firefox, u"io.gitlab.librewolf-community"_s, u"/.librewolf/native-messaging-hosts"_s},
+            {BrowserBase::Chrome, u"com.google.Chrome"_s, u"/config/google-chrome/NativeMessagingHosts"_s},
+            {BrowserBase::Chrome, u"com.google.ChromeDev"_s, u"/config/google-chrome-unstable/NativeMessagingHosts"_s},
+            {BrowserBase::Chromium, u"org.chromium.Chromium"_s, u"/config/chromium/NativeMessagingHosts"_s},
+            {BrowserBase::Chromium, u"io.github.ungoogled_software.ungoogled_chromium"_s, u"/config/chromium/NativeMessagingHosts"_s},
+        };
+
+        // Set up Flatpak permissions for each browser
+        for (const auto &browser : supportedBrowsers) {
+            auto flatpak = new QProcess(this);
+            connect(flatpak, qOverload<int, QProcess::ExitStatus>(&QProcess::finished), this, [flatpak](int, QProcess::ExitStatus) {
+                flatpak->deleteLater();
+            });
+            flatpak->start(u"flatpak"_s, {u"override"_s, u"--user"_s, u"--talk-name=org.kde.plasma.browser.integration"_s, browser.id});
+        }
 
         // Register on separate bus connection to avoid exposing other services to the sandbox
         auto connection = QDBusConnection::connectToBus(QDBusConnection::SessionBus, "org.kde.plasma.browser.integration"_L1);
         connection.registerService("org.kde.plasma.browser.integration"_L1);
         connection.registerObject("/org/kde/plasma/browser/integration"_L1, this, QDBusConnection::ExportAllSlots);
 
-        createMessagingHost(); // always create integration regardless of firefox being installed so we can be ready for the browser
+        // Create messaging hosts for each supported browser
+        for (const auto &browser : supportedBrowsers) {
+            createMessagingHost(browser); // always create integration regardless of the browser being installed so we can be ready for if it is installed
+        }
     }
 
 public Q_SLOTS:
@@ -144,7 +179,14 @@ private:
     {
         constexpr auto maxBufferSize = 1024;
         thread_local std::array<char, maxBufferSize> buffer;
+#ifdef STRERROR_R_CHAR_P
         return strerror_r(error, buffer.data(), buffer.size());
+#else
+        // Won't be changed by strerror_r but not const so compiler doesn't throw an error
+        static char unknown[] = "unknown error";
+
+        return strerror_r(error, buffer.data(), buffer.size()) ? unknown : buffer.data();
+#endif
     }
 
     int openNoSymlinks(const std::filesystem::path &path, int flags, mode_t mode = 0)
@@ -166,28 +208,32 @@ private:
         return openat(dirfd, ".", flags | O_NOFOLLOW | O_CLOEXEC, mode);
     }
 
-    void createMessagingHost()
+    void createMessagingHost(const BrowserInfo &browser)
     {
-        QDir().mkpath(m_hostWrapperDir);
+        const QString hostWrapperDir = QDir::homePath() + QStringLiteral("/.var/app/") + browser.id;
+        const QString hostWrapperName = "plasma-browser-integration-host"_L1;
+        const QString hostWrapperPath = hostWrapperDir + "/"_L1 + hostWrapperName;
 
-        auto hostWrapperDirFd = openNoSymlinks(qUtf8Printable(m_hostWrapperDir), O_PATH);
+        QDir().mkpath(hostWrapperDir);
+
+        auto hostWrapperDirFd = openNoSymlinks(qUtf8Printable(hostWrapperDir), O_PATH);
         const auto closeHostWrapperDirFd = qScopeGuard([hostWrapperDirFd]() {
             close(hostWrapperDirFd);
         });
         if (hostWrapperDirFd == -1) {
             auto err = errno;
-            qCWarning(INTEGRATOR) << "Failed to open hostWrapper directory." << m_hostWrapperDir << ":" << safe_strerror(err);
+            qCWarning(INTEGRATOR) << "Failed to open hostWrapper directory." << hostWrapperDir << ":" << safe_strerror(err);
             return;
         }
 
         { // host wrapper
-            auto hostWrapperFd = openat(hostWrapperDirFd, qUtf8Printable(m_hostWrapperName), O_WRONLY | O_CLOEXEC | O_CREAT | O_TRUNC | O_NOFOLLOW, S_IRWXU);
+            auto hostWrapperFd = openat(hostWrapperDirFd, qUtf8Printable(hostWrapperName), O_WRONLY | O_CLOEXEC | O_CREAT | O_TRUNC | O_NOFOLLOW, S_IRWXU);
             auto closeHostWrapperFd = qScopeGuard([hostWrapperFd]() {
                 close(hostWrapperFd);
             });
             if (hostWrapperFd == -1) {
                 auto err = errno;
-                qCWarning(INTEGRATOR) << "Failed to open host wrapper file" << m_hostWrapperName << ":" << safe_strerror(err);
+                qCWarning(INTEGRATOR) << "Failed to open host wrapper file" << hostWrapperName << ":" << safe_strerror(err);
                 return;
             }
 
@@ -215,7 +261,7 @@ private:
         }
 
         { // hosts definition
-            const QString extensionDefinitionDir = m_hostWrapperDir + "/.mozilla/native-messaging-hosts"_L1;
+            const QString extensionDefinitionDir = hostWrapperDir + browser.nativeMessagingHostsDir;
             QDir().mkpath(extensionDefinitionDir);
 
             auto defintionsDirFd = openNoSymlinks(qUtf8Printable(extensionDefinitionDir), O_PATH);
@@ -244,20 +290,28 @@ private:
                 qCWarning(INTEGRATOR) << "Failed to open extension definition file";
                 return;
             }
+
             QJsonObject extensionDefinitionObject({
                 {u"name"_s, u"org.kde.plasma.browser_integration"_s},
                 {u"description"_s, u"Native connector for KDE Plasma"_s},
-                {u"path"_s, m_hostWrapperPath},
+                {u"path"_s, hostWrapperPath},
                 {u"type"_s, u"stdio"_s},
-                {u"allowed_extensions"_s, QJsonArray({u"plasma-browser-integration@kde.org"_s})},
             });
+
+            // Add browser-specific fields
+            if (browser.base == BrowserBase::Firefox) {
+                extensionDefinitionObject.insert(u"allowed_extensions"_s, QJsonArray({u"plasma-browser-integration@kde.org"_s}));
+            } else if (browser.base == BrowserBase::Chrome || browser.base == BrowserBase::Chromium) {
+                extensionDefinitionObject.insert(u"allowed_origins"_s,
+                                                 QJsonArray({
+                                                     u"chrome-extension://cimiefiiaegbelhefglklhhakcgmhkai/"_s,
+                                                     u"chrome-extension://dnnckbejblnejeabhcmhklcaljjpdjeh/"_s,
+                                                 }));
+            }
+
             extensionDefinition.write(QJsonDocument(extensionDefinitionObject).toJson());
         }
     }
-
-    const QString m_hostWrapperDir = QDir::homePath() + "/.var/app/org.mozilla.firefox"_L1;
-    const QString m_hostWrapperName = "plasma-browser-integration-host"_L1;
-    const QString m_hostWrapperPath = m_hostWrapperDir + "/"_L1 + m_hostWrapperName;
 };
 
 K_PLUGIN_FACTORY_WITH_JSON(BrowserIntegrationFlatpakIntegratorFactory,
