@@ -10,8 +10,11 @@
 #include <QCoreApplication>
 #include <QDBusConnection>
 #include <QDBusObjectPath>
+#include <QDir>
 #include <QGuiApplication>
 #include <QImageReader>
+#include <QMimeDatabase>
+#include <QMimeType>
 
 #include "mprisplayer.h"
 #include "mprisroot.h"
@@ -58,9 +61,24 @@ MPrisPlugin::MPrisPlugin(QObject *parent)
     };
 }
 
+bool MPrisPlugin::onLoad()
+{
+    // This also effectively tells the extension that our host is able to understand album art being downloaded on the extension side.
+    const QList<QByteArray> imageReaderMimeTypes = QImageReader::supportedMimeTypes();
+    QJsonArray mimeTypes;
+    for (const auto &mimeType : imageReaderMimeTypes) {
+        mimeTypes.append(QString::fromLatin1(mimeType));
+    }
+
+    sendData(QStringLiteral("supportedImageMimeTypes"), QJsonObject{{QStringLiteral("mimeTypes"), mimeTypes}});
+    return true;
+}
+
 bool MPrisPlugin::onUnload()
 {
     unregisterService();
+    m_localArtwork.reset();
+    m_artworkUrl.clear();
     return true;
 }
 
@@ -151,24 +169,54 @@ void MPrisPlugin::handleData(const QString &event, const QJsonObject &data)
         m_mediaSrc.clear();
         m_title.clear();
         m_artist.clear();
+
+        m_localArtwork.reset();
         m_artworkUrl.clear();
+        m_pendingArtworkUrl.clear();
+        m_mediaSessionArtworkUrl.clear();
+        m_posterUrl.clear();
+
         m_volume = 1.0;
         m_muted = false;
         m_length = 0;
         m_position = 0;
     } else if (event == QLatin1String("playing")) {
+        bool metadataChanged = false;
+
         setPlaybackStatus(QStringLiteral("Playing"));
 
+        const QString oldEffectiveTitle = effectiveTitle();
+
         m_pageTitle = data.value(QStringLiteral("pageTitle")).toString();
         m_tabTitle = data.value(QStringLiteral("tabTitle")).toString();
 
-        m_url = QUrl(data.value(QStringLiteral("url")).toString());
-        m_mediaSrc = QUrl(data.value(QStringLiteral("mediaSrc")).toString());
+        if (effectiveTitle() != oldEffectiveTitle) {
+            metadataChanged = true;
+        }
 
-        const QUrl posterUrl = QUrl(data.value(QStringLiteral("poster")).toString());
-        if (m_posterUrl != posterUrl) {
-            m_posterUrl = posterUrl;
-            emitPropertyChange(m_player, "Metadata");
+        const QUrl url = QUrl(data.value(QStringLiteral("url")).toString());
+        if (m_url != url) {
+            m_url = url;
+            metadataChanged = true;
+        }
+        const QUrl mediaSrc = QUrl(data.value(QStringLiteral("mediaSrc")).toString());
+        if (m_mediaSrc != mediaSrc) {
+            m_mediaSrc = mediaSrc;
+            metadataChanged = true;
+        }
+
+        // Extension will download artwork.
+        // It will always send this key, even if empty, to tell us to ignore legacy poster/metadata.
+        const QJsonValue artworkPendingValue = data.value(QStringLiteral("pendingArtwork"));
+        const bool artworkPending = artworkPendingValue.isString();
+        m_pendingArtworkUrl = QUrl(artworkPendingValue.toString());
+
+        if (!artworkPending) {
+            const QUrl posterUrl = QUrl(data.value(QStringLiteral("poster")).toString());
+            if (m_posterUrl != posterUrl) {
+                m_posterUrl = posterUrl;
+                metadataChanged = true;
+            }
         }
 
         const qreal oldVolume = volume();
@@ -182,7 +230,7 @@ void MPrisPlugin::handleData(const QString &event, const QJsonObject &data)
 
         const qreal length = data.value(QStringLiteral("duration")).toDouble();
         // <video> duration is in seconds, mpris uses microseconds
-        setLength(length * 1000 * 1000);
+        metadataChanged |= setLength(length * 1000 * 1000);
 
         const qreal position = data.value(QStringLiteral("currentTime")).toDouble();
         setPosition(position * 1000 * 1000);
@@ -214,9 +262,13 @@ void MPrisPlugin::handleData(const QString &event, const QJsonObject &data)
             emitPropertyChange(m_root, "CanSetFullscreen");
         }
 
-        processMetadata(data.value(QStringLiteral("metadata")).toObject()); // also emits metadataChanged signal
+        metadataChanged |= processMetadata(data, !artworkPending);
         processCallbacks(data.value(QStringLiteral("callbacks")).toArray());
 
+        if (metadataChanged) {
+            emitPropertyChange(m_player, "Metadata");
+        }
+
         registerService();
     } else if (event == QLatin1String("paused")) {
         setPlaybackStatus(QStringLiteral("Paused"));
@@ -254,9 +306,21 @@ void MPrisPlugin::handleData(const QString &event, const QJsonObject &data)
         m_muted = data.value(QStringLiteral("muted")).toBool();
         emitPropertyChange(m_player, "Volume");
     } else if (event == QLatin1String("metadata")) {
-        processMetadata(data.value(QStringLiteral("metadata")).toObject());
+        // Extension will download artwork.
+        const QJsonValue artworkPendingValue = data.value(QStringLiteral("pendingArtwork"));
+        const bool artworkPending = artworkPendingValue.isString();
+        m_pendingArtworkUrl = QUrl(artworkPendingValue.toString());
+
+        if (processMetadata(data, !artworkPending)) {
+            emitPropertyChange(m_player, "Metadata");
+        }
     } else if (event == QLatin1String("callbacks")) {
         processCallbacks(data.value(QStringLiteral("callbacks")).toArray());
+    } else if (event == QLatin1String("artwork")) {
+        // Artwork was downloaded by the extension.
+        if (processFetchedArtwork(data)) {
+            emitPropertyChange(m_player, "Metadata");
+        }
     } else if (event == QLatin1String("titlechange")) {
         const QString oldTitle = effectiveTitle();
         m_pageTitle = data.value(QStringLiteral("pageTitle")).toString();
@@ -461,11 +525,17 @@ QVariantMap MPrisPlugin::metadata() const
         metadata.insert(QStringLiteral("xesam:artist"), QStringList{sanitizedString(m_artist)});
     }
 
-    QUrl artUrl = m_artworkUrl;
-    if (!artUrl.isValid()) {
-        artUrl = m_posterUrl;
+    QString artUrlStr;
+    if (m_localArtwork && !m_localArtwork->fileName().isEmpty()) {
+        artUrlStr = QUrl::fromLocalFile(m_localArtwork->fileName()).toString();
     }
-    if (const QString artUrlStr = sanitizedUrlDisplayString(artUrl); !artUrlStr.isEmpty()) {
+    if (artUrlStr.isEmpty()) {
+        artUrlStr = sanitizedUrlDisplayString(m_mediaSessionArtworkUrl);
+    }
+    if (artUrlStr.isEmpty()) {
+        artUrlStr = sanitizedUrlDisplayString(m_posterUrl);
+    }
+    if (!artUrlStr.isEmpty()) {
         metadata.insert(QStringLiteral("mpris:artUrl"), artUrlStr);
     }
 
@@ -492,7 +562,7 @@ void MPrisPlugin::setPlaybackStatus(const QString &playbackStatus)
     }
 }
 
-void MPrisPlugin::setLength(qlonglong length)
+bool MPrisPlugin::setLength(qlonglong length)
 {
     if (m_length != length) {
         const bool oldCanSeek = canSeek();
@@ -503,7 +573,10 @@ void MPrisPlugin::setLength(qlonglong length)
         if (oldCanSeek != canSeek()) {
             emitPropertyChange(m_player, "CanSeek");
         }
+
+        return true;
     }
+    return false;
 }
 
 void MPrisPlugin::setPosition(qlonglong position)
@@ -515,17 +588,38 @@ void MPrisPlugin::setPosition(qlonglong position)
     }
 }
 
-void MPrisPlugin::processMetadata(const QJsonObject &data)
+bool MPrisPlugin::processMetadata(const QJsonObject &data, bool processArtwork)
 {
-    m_title = data.value(QStringLiteral("title")).toString();
-    m_artist = data.value(QStringLiteral("artist")).toString();
-    m_album = data.value(QStringLiteral("album")).toString();
+    bool changed = false;
+
+    const QJsonObject metadata = data.value(QStringLiteral("metadata")).toObject();
+
+    const QString title = metadata.value(QStringLiteral("title")).toString();
+    if (m_title != title) {
+        m_title = title;
+        changed = true;
+    }
+
+    const QString artist = metadata.value(QStringLiteral("artist")).toString();
+    if (m_artist != artist) {
+        m_artist = artist;
+        changed = true;
+    }
+
+    const QString album = metadata.value(QStringLiteral("album")).toString();
+    if (m_album != album) {
+        m_album = album;
+        changed = true;
+    }
+
+    if (!processArtwork) {
+        return changed;
+    }
 
     // for simplicity we just use the biggest artwork it offers, perhaps we could limit it to some extent
-    // TODO download/cache artwork somewhere
     QSize biggest;
     QUrl artworkUrl;
-    const QJsonArray &artwork = data.value(QStringLiteral("artwork")).toArray();
+    const QJsonArray &artwork = metadata.value(QStringLiteral("artwork")).toArray();
 
     const auto &supportedImageMimes = QImageReader::supportedMimeTypes();
 
@@ -566,9 +660,12 @@ void MPrisPlugin::processMetadata(const QJsonObject &data)
         }
     }
 
-    m_artworkUrl = artworkUrl;
+    if (m_mediaSessionArtworkUrl != artworkUrl) {
+        m_mediaSessionArtworkUrl = artworkUrl;
+        changed = true;
+    }
 
-    emitPropertyChange(m_player, "Metadata");
+    return changed;
 }
 
 void MPrisPlugin::processCallbacks(const QJsonArray &data)
@@ -586,6 +683,63 @@ void MPrisPlugin::processCallbacks(const QJsonArray &data)
     }
 }
 
+bool MPrisPlugin::processFetchedArtwork(const QJsonObject &payload)
+{
+    const QUrl artworkUrl(payload.value(QStringLiteral("src")).toString());
+    if (m_pendingArtworkUrl != artworkUrl) {
+        // Website has different metadata by the time the download finished.
+        return false;
+    }
+
+    if (m_artworkUrl == artworkUrl) {
+        // Nothing to do.
+        return false;
+    }
+
+    m_localArtwork.reset();
+    m_artworkUrl.clear();
+    m_pendingArtworkUrl.clear();
+
+    const QByteArray data = dataFromDataUrl(payload.value(QStringLiteral("dataUrl")).toString());
+    if (data.isEmpty()) {
+        // Some error, clear artwork.
+        return true;
+    }
+
+    // For completeness' sake, add a file extension.
+    // We cannot really trust the mimeType provided by the website (YouTube claims it's JPEG but it's actually AVIF)
+    // but it doesn't really matter.
+    QMimeDatabase mimeDb;
+    QMimeType mimeType = mimeDb.mimeTypeForName(payload.value(QStringLiteral("mimeType")).toString());
+    if (!mimeType.isValid() || mimeType.isDefault()) {
+        mimeType = mimeDb.mimeTypeForFileNameAndData(artworkUrl.fileName(), data);
+    }
+
+    QString pattern;
+    pattern.append(QStringLiteral("plasma-browser-integration_artwork_XXXXXX"));
+    if (const QString suffix = mimeType.preferredSuffix(); !suffix.isEmpty()) {
+        pattern.append(QLatin1Char('.'));
+        pattern.append(suffix);
+    }
+
+    const QString filePath = QDir::tempPath() + QLatin1Char('/') + pattern;
+    m_localArtwork = std::make_unique<QTemporaryFile>(filePath);
+    if (!m_localArtwork->open()) {
+        qWarning() << "Failed to open artwork temp file" << filePath << "for writing" << m_localArtwork->errorString();
+        // Local failure, clear artwork.
+        return true;
+    }
+
+    m_localArtwork->write(data);
+
+    // Close and flush before we announce the file via MPRIS.
+    // The file remains until the QTemporaryFile is destroyed.
+    m_localArtwork->close();
+
+    m_artworkUrl = artworkUrl;
+    return true;
+}
+
 void MPrisPlugin::Raise()
 {
     sendData(QStringLiteral("raise"));
